package com.example.repository;

import com.example.domain.Owner;
import jakarta.inject.Singleton;
import org.hibernate.SessionFactory;
import org.hibernate.reactive.stage.Stage;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.function.Function;

@Singleton
public class OwnerRepository {

    private final Stage.SessionFactory sessionFactory;

    OwnerRepository(SessionFactory sessionFactory) {
        this.sessionFactory = sessionFactory.unwrap(Stage.SessionFactory.class);
    }

    public Mono<Void> destroy() {
        return findAll().flatMap(it -> delete(it)).then();
    }

    public Owner create() {
        return new Owner();
    }

    public Mono<Void> save(Owner entity) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.persist(entity)));
    }

    public Mono<Void> delete(Owner entity) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.remove(entity)));
    }

    public Mono<? extends Owner> findById(Long id) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.find(Owner.class, id)));
    }

    public Mono<? extends Owner> findByName(String name) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.createQuery("from Owner where name = :name", Owner.class)
                .setParameter("name", name)
                .getResultList())
            .map(results -> results.iterator().next()));
    }

    public Flux<? extends Owner> findAll() {
        return withSessionFlux(session -> Mono.fromCompletionStage(() -> session.createQuery("from Owner", Owner.class).getResultList())
            .flatMapMany(Flux::fromIterable));
    }

    private <T> Mono<T> withSessionMono(Function<Stage.Session, Mono<T>> consumer) {
        return Mono.deferContextual(contextView -> {
            Stage.Session session = contextView.getOrDefault(TxManager.SESSION_KEY, null);
            if (session == null) {
                return Mono.fromCompletionStage(() -> sessionFactory.withSession(s -> consumer.apply(s).toFuture()));
            }
            return consumer.apply(session);
        });
    }

    private <T> Flux<T> withSessionFlux(Function<Stage.Session, Flux<T>> consumer) {
        return Flux.deferContextual(contextView -> {
            Stage.Session session = contextView.getOrDefault(TxManager.SESSION_KEY, null);
            if (session == null) {
                return Mono.fromCompletionStage(() -> sessionFactory.withSession(s -> consumer.apply(s).collectList().toFuture()))
                    .flatMapMany(Flux::fromIterable);
            }
            return consumer.apply(session);
        });
    }
}
