package com.example.repository;

import com.example.domain.Pet;
import jakarta.inject.Singleton;
import org.hibernate.Hibernate;
import org.hibernate.SessionFactory;
import org.hibernate.reactive.stage.Stage;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.function.Function;

@Singleton
public class PetRepository {

    private final Stage.SessionFactory sessionFactory;

    PetRepository(SessionFactory sessionFactory) {
        this.sessionFactory = sessionFactory.unwrap(Stage.SessionFactory.class);
    }

    public Mono<Void> destroy() {
        return findAll().flatMap(it -> delete(it)).then();
    }

    public Pet create() {
        return new Pet();
    }

    public Mono<Void> save(Pet entity) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.persist(entity)));
    }

    public Mono<Void> delete(Pet entity) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.remove(entity)));
    }

    public Mono<? extends Pet> findByName(String name) {
        return withSessionMono(session -> Mono.fromCompletionStage(() -> session.createQuery("from Pet where name = :name", Pet.class)
                .setParameter("name", name)
                .getResultList())
            .map(results -> results.iterator().next())
            .flatMap(pet -> {
                if (Hibernate.isInitialized(pet.getOwner())) {
                    throw new IllegalStateException("Expected not initialized owner ref");
                }
                // HR000037: Reactive sessions do not support transparent lazy fetching
                // - use Session.fetch() (entity 'example.domain.Owner' with id '1' was not loaded)
                return Mono.fromCompletionStage(() -> session.fetch(pet.getOwner())).map(owner -> {
                    pet.setOwner(owner);
                    if (!Hibernate.isInitialized(pet.getOwner())) {
                        throw new IllegalStateException("Expected initialized owner ref");
                    }
                    return pet;
                });
            }));
    }

    public Flux<Pet> findAll() {
        return withSessionFlux(session -> Mono.fromCompletionStage(() -> session.createQuery("from Pet", Pet.class)
                .getResultList()
                .thenCompose(pets -> {
                    // HR000037: Reactive sessions do not support transparent lazy fetching
                    // - use Session.fetch() (entity 'example.domain.Owner' with id '1' was not loaded)
                    CompletableFuture<List<Pet>> resultStage = CompletableFuture.completedFuture(new ArrayList<>());
                    for (Pet pet : pets) {
                        resultStage = resultStage
                            .thenCompose(newPets -> session
                                .fetch(pet.getOwner())
                                .thenApply(owner -> {
                                    pet.setOwner(owner);
                                    newPets.add(pet);
                                    return newPets;
                                })
                            );
                    }
                    return resultStage;
                }))
            .flatMapMany(Flux::fromIterable));
    }

    private <T> Mono<T> withSessionMono(Function<Stage.Session, Mono<T>> consumer) {
        return Mono.deferContextual(contextView -> {
            Stage.Session session = contextView.getOrDefault(TxManager.SESSION_KEY, null);
            if (session == null) {
                return Mono.fromCompletionStage(() -> sessionFactory.withSession(s -> consumer.apply(s).toFuture()));
            }
            return consumer.apply(session);
        });
    }

    private <T> Flux<T> withSessionFlux(Function<Stage.Session, Flux<T>> consumer) {
        return Flux.deferContextual(contextView -> {
            Stage.Session session = contextView.getOrDefault(TxManager.SESSION_KEY, null);
            if (session == null) {
                return Mono.fromCompletionStage(() -> sessionFactory.withSession(s -> consumer.apply(s).collectList().toFuture()))
                    .flatMapMany(Flux::fromIterable);
            }
            return consumer.apply(session);
        });
    }
}
